"""Transform an UFL expression into list of accumulation terms."""

from dune.perftool.ufl.flatoperators import construct_binary_operator
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.transformations.replace import replace_expression
from dune.perftool.ufl.transformations.identitypropagation import identity_propagation
from dune.perftool.ufl.transformations.order_gradient_restriction import order_gradient_restriction
from dune.perftool.ufl.transformations.zeropropagation import zero_propagation
from dune.perftool.ufl.modified_terminals import ModifiedArgument
from dune.perftool.pdelab.restriction import Restriction
from dune.perftool.options import get_option

from ufl.classes import Zero, Identity, Indexed, IntValue, MultiIndex, Product, IndexSum
from ufl.core.multiindex import indices

from pytools import Record


class AccumulationTerm(Record):
    """Store informations about accumulation terms

    Arguments:
    ----------
    term: The UFL expression we want to accumulate
    argument: Corresponding test function (in case we split it) without indices
    new_indices: Indices of the test function
    """
    def __init__(self,
                 term,
                 argument=None,
                 new_indices=None
                 ):
        assert (isinstance(argument, ModifiedArgument) or argument is None)

        # Test if this expression belongs to a jacobian method
        all_jacobian_args = extract_modified_arguments(term,
                                                       argnumber=1,
                                                       do_index=False,
                                                       do_gradient=False)
        is_jacobian = False
        if all_jacobian_args:
            is_jacobian = True

        Record.__init__(self,
                        term=term,
                        argument=argument,
                        new_indices=new_indices,
                        is_jacobian=is_jacobian
                        )

    def indexed_test_arg(self):
        """Return test argument of this accumulation term with its indices"""
        if self.new_indices is None:
            return self.argument.expr
        else:
            return Indexed(self.argument.expr, MultiIndex(self.new_indices))


def split_into_accumulation_terms(expr, cut_test_arg=False, split_gradients=False):
    """Split an expression into a list of AccumulationTerms

    Arguments:
    ----------
    expr: The UFL expression we split into AccumulationTerms
    cut_test_arg: Wheter we want to return AccumulationTerms where the
        test function is cut from the rest of the expression.
    """
    expr = order_gradient_restriction(expr)

    expression_list = split_expression(expr, split_gradients=split_gradients)
    acc_term_list = []

    for e in expression_list:
        if cut_test_arg:
            acc_term_list.append(cut_accumulation_term(e))
        else:
            acc_term_list.append(AccumulationTerm(e, ModifiedArgument(expr=None)))
    return acc_term_list


@ufl_transformation(name="splitt_expression", extraction_lambda=lambda l: [e for e in l])
def split_expression(expr, split_gradients=False):
    """Split expression into a list of expressions

    Note: This is not an ufl transformation since it does not return
    an expression. It is nonetheless decorated with ufl_transformation
    since the decorator can handle lists of expressions and it can be
    useful for debugging to get the corresponding trees.
    """
    # Store AccumulationTerms in this list
    ret = []

    # Extract a list of modified terminals for the test function. We
    # will split the expression into one part for each moidified argument.
    test_args = extract_modified_arguments(expr,
                                           argnumber=0,
                                           do_index=False,
                                           do_gradient=split_gradients)

    for test_arg in test_args:
        # Do this as a multi step replacement procedure to avoid UFL
        # nagging about too much stuff in reconstructing the
        # expressions

        # 1) We first cut the expression to the relevant modified test_function
        # by replacing all other test functions with zero
        replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                     free_indices=ma.expr.ufl_free_indices,
                                     index_dimensions=ma.expr.ufl_index_dimensions)
                       for ma in test_args}
        replacement[test_arg.expr] = test_arg.expr
        replace_expr = replace_expression(expr, replacemap=replacement)

        # 2) Propagate indexed zeros to simplify expression
        replace_expr = zero_propagation(replace_expr)

        # Test if we have a jacobian by extracting aguments with argnumber=1
        all_jacobian_args = extract_modified_arguments(expr,
                                                       argnumber=1,
                                                       do_index=False,
                                                       do_gradient=split_gradients)

        # 3) Further split according to trial function in jacobian terms
        #
        # Note: We need to split according to the FunctionView. For
        # example in Stokes with test functions v and q we have to
        # split between those.
        #
        # But: We don't want to split according to gradients since
        # this would break the input buffers of stage 3 of
        # sumfactorization
        if all_jacobian_args:
            trial_args = extract_modified_arguments(replace_expr,
                                                    argnumber=1,
                                                    do_index=False,
                                                    do_gradient=False)
            for trial_arg in trial_args:
                # 3.1) Restrict to this trial argument
                replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                             free_indices=ma.expr.ufl_free_indices,
                                             index_dimensions=ma.expr.ufl_index_dimensions)
                               for ma in trial_args}
                replacement[trial_arg.expr] = trial_arg.expr
                jac_expr = replace_expression(replace_expr, replacemap=replacement)

                # 3.2) Propagate indexed zeros to simplify expression
                jac_expr = zero_propagation(jac_expr)

                # 3.3) Split according to restriction
                indexed_jac_args = extract_modified_arguments(jac_expr, argnumber=1, do_index=True, do_gradient=False)
                for restriction in (Restriction.NONE, Restriction.POSITIVE, Restriction.NEGATIVE):
                    replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                                 free_indices=ma.expr.ufl_free_indices,
                                                 index_dimensions=ma.expr.ufl_index_dimensions)
                                   if ma.restriction != restriction else ma.expr
                                   for ma in indexed_jac_args}

                    acc_expr = replace_expression(jac_expr, replacemap=replacement)

                    if not isinstance(acc_expr, Zero):
                        ret.append(acc_expr)
        else:
            if not isinstance(replace_expr, Zero):
                ret.append(replace_expr)

    return ret


@ufl_transformation(name="cut_accum", extraction_lambda=lambda l: [l.term])
def cut_accumulation_term(expr):
    """Cut test function from expression and return AccumulationTerm

    Note: This assumes that there is only one test function in the
    expression. You need to make sure to split your expression into
    appropriate parts before calling this!
    """
    # If there are multiple test arguments or gradients and
    # non-gradients something wen wrong!
    test_args = extract_modified_arguments(expr, argnumber=0, do_index=False, do_gradient=True)
    assert len(test_args) == 1
    test_arg = test_args[0]

    # 1) Cut the test function itself from the expression
    #
    # This is done by replacing the test function with an
    # appropriate product of identity matrices. This way we can
    # make sure that the indices of the result will be right. This
    # is best explained by an example:
    #
    # Suppose we have the following expression:
    #
    # \sum_{i,j} a_{i,j} (\nabla v)_{i,j} + \sum_{k,l} b_{k,l} (\nable v)_{k,l}
    #
    # If we want to cut the gradient of the test function v we
    # need to make sure, that both sums have the right indices:
    #
    # \sum_{m,n} (a_{m,n} + b_{m,n}) (\nabla v)_{m,n}
    #
    # and we extract (a_{m,n} + b_{m,n}). We achieve that by the
    # following replacements:
    #
    # (\nabla v)_{i,j} -> I_{m,i} I_{n,j}
    # (\nabla v)_{k,l} -> I_{m,k} I_{n,l}
    #
    # Resulting in:
    #
    # \sum_{i,j} a_{i,j} I_{m,i} I_{n,j} + \sum_{k,l} b_{k,l} I_{m,k} I_{n,l}
    #
    # In step 2 this will collaps to: a_{m,n} + b_{m,n}
    replacement = {}
    newi = None
    backmap = {}

    if get_option("unroll_dimension_loops"):
        # This code path is much simpler than the one below: The modified arguments
        # of the test function cannot use free indices!
        indexed_test_args = extract_modified_arguments(expr, argnumber=0)
        assert len(indexed_test_args) == 1
        ma, = indexed_test_args
        replacement[ma.expr] = IntValue(1)
    else:
        # Get all appearances of the test function with their indices
        indexed_test_args = extract_modified_arguments(expr, argnumber=0, do_index=True)
        for indexed_test_arg in indexed_test_args:
            if indexed_test_arg.index:
                # If the test function is indexed, create a new multiindex of this shape
                # -> (m,n) in the example above
                if newi is None:
                    newi = indices(len(indexed_test_arg.index))

                # This handles the special case with two identical
                # indices on an test function. E.g. in Stokes on an
                # axiparallel grid you get a term:
                #
                # -(\sum_i K_{i,i} (\nabla v)_{i,i}) w
                #   = \sum_k \sum_l (-K_{k,k} w I_{k,l} (\nabla v)_{k,l})
                #
                # and we want to split
                #
                # -K_{k,k} w I_{k,l} corresponding to (\nabla v)_{k,l}.
                #
                # This is done by:
                # - Replacing (\nabla v)_{i,i} with I_{k,i}*(\nabla
                #   v)_{k,l}. Here (\nabla v)_{k,l} serves as a
                #   placeholder and will be replaced later on.
                # - Propagating the identity in step 4.
                # - Replacing (\nabla v)_{k,l} by I_{k,l} after step 4.
                if len(set(indexed_test_arg.index._indices)) < len(indexed_test_arg.index._indices):
                    if len(indexed_test_arg.index._indices) > 2:
                        raise NotImplementedError("Test argument with more than three indices and double occurence ist not implemented.")
                    mod_index_map = {indexed_test_arg.index: MultiIndex((newi[0], newi[1]))}
                    mod_indexed_test_arg = replace_expression(indexed_test_arg.expr,
                                                              replacemap=mod_index_map)
                    rep = Product(Indexed(Identity(2),
                                          MultiIndex((newi[0], indexed_test_arg.index[0]))),
                                  mod_indexed_test_arg)
                    backmap.update({mod_indexed_test_arg:
                                    Indexed(Identity(2), MultiIndex((newi[0], newi[1])))})
                    replacement.update({indexed_test_arg.expr: rep})
                else:
                    # Replace indexed test function with a product of identities.
                    identities = tuple(Indexed(Identity(2), MultiIndex((i,) + (j,)))
                                       for i, j in zip(newi, indexed_test_arg.index._indices))
                    replacement.update({indexed_test_arg.expr:
                                        construct_binary_operator(identities, Product)})
            else:
                replacement.update({indexed_test_arg.expr: IntValue(1)})

    expr = replace_expression(expr, replacemap=replacement)

    # 2) Collapse any identity nodes that may have been introduced
    # by replacing vectors and maybe replace placeholder from last step
    expr = identity_propagation(expr)
    expr = replace_expression(expr, replacemap=backmap)

    return AccumulationTerm(expr, test_arg, newi)
